      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
         
         ! Uncomment these lines if you wish to use the functions in this file,
         ! otherwise we use a null_ version which does nothing.
         s% extras_startup => extras_startup
         s% extras_check_model => extras_check_model
         s% extras_finish_step => extras_finish_step
         s% extras_after_evolve => extras_after_evolve
         s% how_many_extra_history_columns => how_many_extra_history_columns
         s% data_for_extra_history_columns => data_for_extra_history_columns
         s% how_many_extra_profile_columns => how_many_extra_profile_columns
         s% data_for_extra_profile_columns => data_for_extra_profile_columns  

         ! Once you have set the function pointers you want,
         ! then uncomment this (or set it in your star_job inlist)
         ! to disable the printed warning message,
          s% job% warn_run_star_extras =.false.       
            
      end subroutine extras_controls
      
      ! None of the following functions are called unless you set their
      ! function point in extras_control.
      
      
      integer function extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_startup = 0
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
      end function extras_startup
      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_check_model = keep_going         
         if (.false. .and. s% star_mass_h1 < 0.35d0) then
            ! stop when star hydrogen mass drops to specified level
            extras_check_model = terminate
            write(*, *) 'have reached desired hydrogen mass'
            return
         end if


         ! if you want to check multiple conditions, it can be useful
         ! to set a different termination code depending on which
         ! condition was triggered.  MESA provides 9 customizeable
         ! termination codes, named t_xtra1 .. t_xtra9.  You can
         ! customize the messages that will be printed upon exit by
         ! setting the corresponding termination_code_str value.
         ! termination_code_str(t_xtra1) = 'my termination condition'

         ! by default, indicate where (in the code) MESA terminated
         if (extras_check_model == terminate) s% termination_code = t_extras_check_model
      end function extras_check_model


      integer function how_many_extra_history_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_columns = 0
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         !note: do NOT add the extras names to history_columns.list
         ! the history_columns.list is only for the built-in log column options.
         ! it must not include the new column names you are adding here.
         

      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(id, id_extra)
         use star_def, only: star_info
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_columns = 0
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
         use star_def, only: star_info, maxlen_profile_column_name
         use const_def, only: dp
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         !note: do NOT add the extra names to profile_columns.list
         ! the profile_columns.list is only for the built-in profile column options.
         ! it must not include the new column names you are adding here.

         ! here is an example for adding a profile column
         !if (n /= 1) stop 'data_for_extra_profile_columns'
         !names(1) = 'beta'
         !do k = 1, nz
         !   vals(k,1) = s% Pgas(k)/s% P(k)
         !end do
         
      end subroutine data_for_extra_profile_columns
      

      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      integer function extras_finish_step(id, id_extra)
         use chem_def

         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going
         call store_extra_info(s)

         ! to save a profile, 
            ! s% need_to_save_profiles_now = .true.
         ! to update the star log,
            ! s% need_to_update_history_now = .true.

         ! Output models at each 0.1 central helium abundance during he core burning
         ! Adding also core oxygen depletion and carbon depletion
         ! YLVA 29/3 - 2016

         s% lxtra1 = .false.    ! for helium core burning, in steps of core mass fraction 0.1
         s% lxtra2 = .false.
         s% lxtra3 = .false.
         s% lxtra4 = .false.
         s% lxtra5 = .false.
         s% lxtra6 = .false.
         s% lxtra7 = .false.
         s% lxtra8 = .false.
         s% lxtra9 = .false.
       
         s% lxtra10 = .false.   ! For core carbon and oxygen depletion
         s% lxtra11 = .false.
         
         s% lxtra12 = .false.   ! for Hertzsprung gap, just when starting hydrogen shell burning

         s% lxtra13 = .false.   ! for the main sequence, h_frac = 0.5

         s% lxtra14 = .false.   ! for carbon core burning, c_frac = 0.2, 0.4, 0.5, 0.6, 0.8
         s% lxtra15 = .false.
         s% lxtra16 = .false.
         s% lxtra17 = .false.
         s% lxtra18 = .false.

         if (s%xa(s%net_iso(ih1), s%nz) .LE. 0.5) then   ! half-way through main sequence
            s% lxtra13 = .true.
            if ((s%lxtra13_old .eqv. .false.) .AND. (s% lxtra13 .eqv. .true.)) then
               s% need_to_save_profiles_now = .true.
               s% need_to_update_history_now = .true.
               s% save_profiles_model_priority = 113
            end if
         end if

         if (s%xa(s%net_iso(ih1), s%nz) .LE. 1d-3) then !TAMS
             if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.1) then  ! Helium left in core 0.1
                s% lxtra1 = .true.
                if ((s%lxtra1_old .eqv. .false.) .AND. (s% lxtra1 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 101 
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.2) then  ! Helium left in core 0.2
                s% lxtra2 = .true.
                if ((s%lxtra2_old .eqv. .false.) .AND. (s% lxtra2 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 102 
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.3) then  ! Helium left in core 0.3
                s% lxtra3 = .true.
                if ((s%lxtra3_old .eqv. .false.) .AND. (s% lxtra3 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 103 
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.4) then  ! Helium left in core 0.4
                s% lxtra4 = .true.
                if ((s%lxtra4_old .eqv. .false.) .AND. (s% lxtra4 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 104
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.5) then  ! Helium left in core 0.5
                s% lxtra5 = .true.
                if ((s%lxtra5_old .eqv. .false.) .AND. (s% lxtra5 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 105
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.6) then  ! Helium left in core 0.6
                s% lxtra6 = .true.
                if ((s%lxtra6_old .eqv. .false.) .AND. (s% lxtra6 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 106
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.7) then  ! Helium left in core 0.7
                s% lxtra7 = .true.
                if ((s%lxtra7_old .eqv. .false.) .AND. (s% lxtra7 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 107
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.8) then  ! Helium left in core 0.8
                s% lxtra8 = .true.
                if ((s%lxtra8_old .eqv. .false.) .AND. (s% lxtra8 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 108
                end if
             else if (s%xa(s%net_iso(ihe4), s%nz) .LE. 0.9) then  ! Helium left in core 0.9
                s% lxtra9 = .true.
                if ((s%lxtra9_old .eqv. .false.) .AND. (s% lxtra9 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 109 
                end if
             end if
             
             ! Here comes later stages (carbon and oxygen depletion)
             if (s%xa(s%net_iso(ihe4), s%nz) .LE. 1d-3) then   ! Helium TAMS

                ! This is for carbon core burning
                if (s%xa(s%net_iso(ic12), s%nz) .LE. 0.2) then
                    s% lxtra14 = .true.
                    if ((s%lxtra14_old .eqv. .false.) .AND. (s% lxtra14 .eqv. .true.)) then
                       s% need_to_save_profiles_now = .true.
                       s% need_to_update_history_now = .true.
                       s% save_profiles_model_priority = 114
                    end if
                else if (s%xa(s%net_iso(ic12), s%nz) .LE. 0.4) then
                    s% lxtra15 = .true.
                    if ((s%lxtra15_old .eqv. .false.) .AND. (s% lxtra15 .eqv. .true.)) then
                       s% need_to_save_profiles_now = .true.
                       s% need_to_update_history_now = .true.
                       s% save_profiles_model_priority = 115
                    end if  
                else if (s%xa(s%net_iso(ic12), s%nz) .LE. 0.5) then
                    s% lxtra16 = .true.
                    if ((s%lxtra16_old .eqv. .false.) .AND. (s% lxtra16 .eqv. .true.)) then
                       s% need_to_save_profiles_now = .true.
                       s% need_to_update_history_now = .true.
                       s% save_profiles_model_priority = 116
                    end if
                else if (s%xa(s%net_iso(ic12), s%nz) .LE. 0.6) then
                    s% lxtra17 = .true.
                    if ((s%lxtra17_old .eqv. .false.) .AND. (s% lxtra17 .eqv. .true.)) then
                       s% need_to_save_profiles_now = .true.
                       s% need_to_update_history_now = .true.
                       s% save_profiles_model_priority = 117
                    end if
               else if (s%xa(s%net_iso(ic12), s%nz) .LE. 0.8) then
                    s% lxtra18 = .true.
                    if ((s%lxtra18_old .eqv. .false.) .AND. (s% lxtra18 .eqv. .true.)) then
                       s% need_to_save_profiles_now = .true.
                       s% need_to_update_history_now = .true.
                       s% save_profiles_model_priority = 118
                    end if
                end if

                if (s%xa(s%net_iso(ic12), s%nz) .LE. 1d-2) then  ! Carbon core depletion (Sukhbold & Woosley 2014, end of 2.1) 
                    s% lxtra10 = .true.
                    if ((s%lxtra10_old .eqv. .false.) .AND. (s% lxtra10 .eqv. .true.)) then
                       s% need_to_save_profiles_now = .true.
                       s% need_to_update_history_now = .true.
                       s% save_profiles_model_priority = 110
                       extras_finish_step = terminate           ! Stop at carbon core depletion !!!!!!!! YLVA 14/4 - 2016
                    end if
                end if
                if (s%xa(s%net_iso(io16), s%nz) .LE. 4d-2) then   ! Oxygen core depletion (Sukhbold & Woosley 2014, 2.1)
                    s% lxtra11 = .true.
                    if ((s%lxtra11_old .eqv. .false.) .AND. (s% lxtra11 .eqv. .true.)) then
                       s% need_to_save_profiles_now = .true.
                       s% need_to_update_history_now = .true.
                       s% save_profiles_model_priority = 111
                    end if
                end if
             end if
       
             ! Here is for right after the TAMS, first model of Hertzsprung gap
             if ((s%L_nuc_burn_total .GE. 0.5*s%L_phot) .AND. (s%xa(s%net_iso(ihe4), s%nz) .GE. 0.98)) then     
                s% lxtra12 = .true.
                if ((s%lxtra12_old .eqv. .false.) .AND. (s% lxtra12 .eqv. .true.)) then
                   s% need_to_save_profiles_now = .true.
                   s% need_to_update_history_now = .true.
                   s% save_profiles_model_priority = 112
                end if
             end if

         end if

         ! see extras_check_model for information about custom termination codes
         ! by default, indicate where (in the code) MESA terminated
         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
      end function extras_finish_step
      
      
      subroutine extras_after_evolve(id, id_extra, ierr)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine extras_after_evolve
      
      
      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      
      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info
